import json
import os
import time
from io import BytesIO

import requests
from PIL import Image
from nostr_sdk import Kind

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.nip88_utils import NIP88Config
from nostr_dvm.utils.nip89_utils import NIP89Config, check_and_set_d_tag
from nostr_dvm.utils.output_utils import upload_media_to_hoster

"""
This File contains a module to transform Text input on stabity AI API

Accepted Inputs: Prompt (text)
Outputs: An url to an Image
"""
#STABILITY_KEY = os.getenv("STABILITY_KEY")



class ImageGenerationSD35(DVMTaskInterface):
    KIND: Kind = EventDefinitions.KIND_NIP90_GENERATE_IMAGE
    TASK: str = "text-to-image"
    FIX_COST: float = 50

    def __init__(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                 admin_config: AdminConfig = None,
                 options=None, task=None):
        super().__init__(name, dvm_config, nip89config, nip88config, admin_config, options, task)
        if options is not None:
            self.STABILITY_KEY = options["API_KEY"]

    async def init_dvm(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                       admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)

    async def is_input_supported(self, tags, client=None, dvm_config=None):
        for tag in tags:
            if tag.as_vec()[0] == 'i':
                input_value = tag.as_vec()[1]
                input_type = tag.as_vec()[2]
                if input_type != "text":
                    return False

            elif tag.as_vec()[0] == 'output':
                output = tag.as_vec()[1]
                if (output == "" or
                        not (output == "image/png" or "image/jpg"
                             or output == "image/png;format=url" or output == "image/jpg;format=url")):
                    print("Output format not supported, skipping..")
                    return False

        return True

    async def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        request_form = {"jobID": event.id().to_hex() + "_" + self.NAME.replace(" ", "")}

        prompt = ""
        negative_prompt = ""
        if self.options.get("default_model") and self.options.get("default_model") != "":
            model = self.options['default_model']
        else:
            model = "sd3" #ultra

        ratio_width = "4"
        ratio_height = "5"
        width = ""
        height = ""
        seed = "0"
        if self.options.get("default_lora") and self.options.get("default_lora") != "":
            lora = self.options['default_lora']
        else:
            lora = ""
        lora_weight = ""
        strength = ""
        guidance_scale = ""
        for tag in event.tags().to_vec():
            if tag.as_vec()[0] == 'i':
                input_type = tag.as_vec()[2]
                if input_type == "text":
                    prompt = tag.as_vec()[1]

            elif tag.as_vec()[0] == 'param':
                print("Param: " + tag.as_vec()[1] + ": " + tag.as_vec()[2])
                if tag.as_vec()[1] == "negative_prompt":
                    negative_prompt = tag.as_vec()[2]
                elif tag.as_vec()[1] == "ratio":
                    if len(tag.as_vec()) > 3:
                        ratio_width = (tag.as_vec()[2])
                        ratio_height = (tag.as_vec()[3])
                    elif len(tag.as_vec()) == 3:
                        split = tag.as_vec()[2].split(":")
                        ratio_width = split[0]
                        ratio_height = split[1]
                    # if size is set it will overwrite ratio.
                elif tag.as_vec()[1] == "size":
                    if len(tag.as_vec()) > 3:
                        width = (tag.as_vec()[2])
                        height = (tag.as_vec()[3])
                    elif len(tag.as_vec()) == 3:
                        split = tag.as_vec()[2].split("x")
                        if len(split) > 1:
                            width = split[0]
                            height = split[1]
                elif tag.as_vec()[1] == "model":
                    model = tag.as_vec()[2]
                elif tag.as_vec()[1] == "seed":
                    seed = tag.as_vec()[2]



        options = {
            "prompt": prompt,
            "negative_prompt": negative_prompt,
            "ratio": ratio_width + ':' + ratio_height,
            "width": width,
            "height": height,
            "seed": seed,
            "model": model,
        }
        request_form['options'] = json.dumps(options)

        return request_form

    def send_async_generation_request(self,
            host,
            params,
    ):

        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self.STABILITY_KEY}",

        }

        # Encode parameters
        files = {}
        if "image" in params:
            image = params.pop("image")
            files = {"image": open(image, 'rb')}

        # Send request
        print(f"Sending REST request to {host}...")
        response = requests.post(
            host,
            headers=headers,
            files=files,
            data=params
        )
        if not response.ok:
            raise Exception(f"HTTP {response.status_code}: {response.text}")

        # Process async response
        response_dict = json.loads(response.text)
        generation_id = response_dict.get("id", None)
        assert generation_id is not None, "Expected id in response"

        # Loop until result or timeout
        timeout = int(os.getenv("WORKER_TIMEOUT", 500))
        start = time.time()
        status_code = 202
        while status_code == 202:
            response = requests.get(
                f"{host}/result/{generation_id}",
                headers={
                    **headers,
                    "Accept": "image/*"
                },
            )

            if not response.ok:
                raise Exception(f"HTTP {response.status_code}: {response.text}")
            status_code = response.status_code
            time.sleep(10)
            if time.time() - start > timeout:
                raise Exception(f"Timeout after {timeout} seconds")

        return response

    def send_generation_request(self,
            host,
            params,
    ):
        headers = {
            "Accept": "image/*",
            "Authorization": f"Bearer {self.STABILITY_KEY}"
        }



        # Encode parameters
        files = {}
        image = params.pop("image", None)
        mask = params.pop("mask", None)
        if image is not None and image != '':
            files["image"] = open(image, 'rb')
        if mask is not None and mask != '':
            files["mask"] = open(mask, 'rb')
        if len(files) == 0:
            files["none"] = ''

        # Send request
        print(f"Sending REST request to {host}...")
        response = requests.post(
            host,
            headers=headers,
            files=files,
            data=params
        )
        if not response.ok:
            raise Exception(f"HTTP {response.status_code}: {response.text}")

        return response

    async def process(self, request_form):
        try:

            options = self.set_options(request_form)


            #host = f"https://api.stability.ai/v2beta/stable-image/generate/ultra"
            #host = f"https://api.stability.ai/v2beta/stable-image/generate/sd3"
            model = "sd3.5-large"

            params = {
                "prompt": options['prompt'], # "dark high contrast render of a psychedelic tree of life illuminating dust in a mystical cave.",
                "negative_prompt": options['negative_prompt'],
                "aspect_ratio": options['ratio'],
                #"seed": int(options['seed']),
                "output_format": "jpeg",
                #"model": model,
                #"mode": "text-to-image"
            }

            host =  f"https://api.stability.ai/v2beta/stable-image/generate/{options['model']}"
            host = f"https://api.stability.ai/v2beta/stable-image/generate/ultra"
            response = self.send_generation_request(host, params)
            image = Image.open(BytesIO(response.content)).convert("RGB")
            image.save("./outputs/image.jpg")
            result = await upload_media_to_hoster("./outputs/image.jpg")
            return result

        except Exception as e:
            raise Exception(e)


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config, server_address, default_model="sd3", default_lora=""):
    dvm_config = build_default_config(identifier)
    dvm_config.USE_OWN_VENV = False
    admin_config.LUD16 = dvm_config.LN_ADDRESS
    # A module might have options it can be initialized with, here we set a default model, and the server
    # address it should use. These parameters can be freely defined in the task component
    options = {'default_model': default_model, 'default_lora': default_lora, 'server': server_address}

    nip89info = {
        "name": name,
        "picture": "https://i.nostr.build/NOXcCIPmOZrDTK35.jpg",
        "about": "I draw images using Stable diffusion ultra",
        "supportsEncryption": True,
        "acceptsNutZaps": dvm_config.ENABLE_NUTZAP,
        "nip90Params":{
            "negative_prompt": {
                "required": False,
                "values": []
            },
            "ratio": {
                "required": False,
                "values": ["1:1", "5:4", "3:2", "16:9", "21:9", "9:21", "9:16", "2:3", "4:5"]
            },
            "seed": {
                "required": False,
                "values": []
            }
        }
    }
    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["picture"])
    nip89config.CONTENT = json.dumps(nip89info)

    return ImageGenerationSD35(name=name, dvm_config=dvm_config, nip89config=nip89config,
                               admin_config=admin_config, options=options)


if __name__ == '__main__':
    process_venv(ImageGenerationSD35)



import asyncio
import hashlib
from datetime import timedelta
from io import BytesIO
from urllib.request import urlopen
import re
import mimetypes
import os


from PIL import Image
import numpy
from nostr_sdk import Keys, EventBuilder, Kind, Client, NostrSigner, Filter, EventId, Tag, PublicKey
from blurhash import  encode
import requests
from urllib.parse import urlparse


from nostr_dvm.utils.nostr_utils import send_event, print_send_result
from nostr_dvm.utils.print_utils import bcolors


async def gallery_announce_list(tags, dvm_config, client):
    keys = Keys.parse(dvm_config.NIP89.PK)
    content = ""
    event = EventBuilder(Kind(10011), content).tags(tags).sign_with_keys(keys)
    eventid = await send_event(event, client=client, dvm_config=dvm_config)

    print(
        bcolors.BLUE + "[" + dvm_config.NIP89.NAME + "] Announced Gallery for " + dvm_config.NIP89.NAME + " (EventID: " + str(
            eventid.to_hex()) + ")" + bcolors.ENDC)


async def convert_nip93_to_nip68(private_key, relay_list, user_to_import_npub=None, startindex=0):
    keys = Keys.parse(private_key)
    if user_to_import_npub is None:
        user_to_import_npub = keys.public_key().to_hex()
    
    client = Client(NostrSigner.keys(keys))
    for relay in relay_list:
        await client.add_relay(relay)
    await client.connect()

    nip93_filter = Filter().kind(Kind(1163)).author(PublicKey.parse(user_to_import_npub))

    events =  await client.fetch_events([nip93_filter], timedelta(5))

    events_vec = events.to_vec()
    reversed_events_vec = reversed(events_vec)
    counter = -1
    for event in reversed_events_vec:
        counter += 1
        if counter < startindex:
            continue

        image_url = ""
        content = ""
        m = ""
        x = ""
        dim = ""
        blurhash = ""
        size = ""
        for tag in event.tags().to_vec():
            if tag.as_vec()[0] == "url":
                image_url = tag.as_vec()[1]
            elif tag.as_vec()[0] == "m":
                m = tag.as_vec()[1]
            elif tag.as_vec()[0] == "x":
                x = tag.as_vec()[1]
            elif tag.as_vec()[0] == "dim":
                dim = tag.as_vec()[1]
            elif tag.as_vec()[0] == "blurhash":
                blurhash = tag.as_vec()[1]
            elif tag.as_vec()[0] == "e":
                eventid = tag.as_vec()[1]
                if len(tag.as_vec()) == 3:
                    relay_hint = tag.as_vec()[2]
                    try:
                        await client.connect_relay(relay_hint)
                    except Exception as e:
                        print(relay_hint)
                e_filter = Filter().id(EventId.parse(eventid)).limit(1)
                content_events = await client.fetch_events([e_filter], timedelta(5))
                if len(content_events.to_vec()) > 0:
                    content_event = content_events.to_vec()[0]
                    content =  re.sub(r'^https?:\/\/.*[\r\n]*', '', content_event.content(), flags=re.MULTILINE).rstrip()

        var = input("Convert and post this image? (y(es)/n(o)/d(elete): " + image_url + " Content: " + content + "\n")
        if var == "d" or var == "delete":
            delete_event = EventBuilder.delete([event.id()]).sign_with_keys(keys)
            response_status = await client.send_event(delete_event)
            print_send_result(response_status)



        elif var == "y" or var == "yes":
            try:
                response = requests.get(image_url)
                img = Image.open(BytesIO(response.content)).convert("RGB")

                if blurhash == "":
                    blurhash = encode(numpy.array(img))
                if dim == "":
                    width, height = img.size
                    dim = str(width) + "x" + str(height)
                if m == "":
                    m, _ = mimetypes.guess_type(image_url)
                if size == "":
                    try:
                        d = urlopen(image_url)
                        size = str(d.info()['Content-Length'])
                    except Exception:
                        size = ""
                if x == "":
                    e = BytesIO(response.content)

                    a = urlparse(image_url)
                    print(os.path.basename(a.path))
                    with open(os.path.basename(a.path), "wb") as f:
                        f.write(e.getbuffer())
                    h = hashlib.sha256()
                    with open(os.path.basename(a.path), 'rb') as file:
                        while True:
                            # Reading is buffered, so we can read smaller chunks.
                            chunk = file.read(h.block_size)
                            if not chunk:
                                break
                            h.update(chunk)
                    x_old = x
                    x = h.hexdigest()

            except Exception as e:
                print(e)


            print("Image: " + image_url + " Content: " + content + " m: " + m + " x: " + x + " dim: " + dim + " size: " + size + " blurhash: " + blurhash)

            imeta_tag = Tag.parse(["imeta", "url " + image_url, "m " + m, "alt " + content, "x " + x, "size " + size, "blurhash " + blurhash  ])
            x_tag = Tag.parse(["x", x])
            alt_tag = Tag.parse(["alt", "List of pictures"])
            m_tag = Tag.parse(["m", m])
            tags = [alt_tag, imeta_tag, x_tag, m_tag]
            nip68event = EventBuilder(Kind(20), content).tags(tags).sign_with_keys(keys)
            print(nip68event)

            await client.send_event(nip68event)




if __name__ == '__main__':
    private_key = "nsec..."

    RELAY_LIST = ["wss://relay.primal.net",
                  "wss://relay.damus.io",
                  "wss://relay.nostrplebs.com",
                  "wss://nostr.mom",
                  "wss://nostr.oxtr.dev",
                  "wss://relay.nostr.band"
                  ]
    startindex = 0
    asyncio.run(convert_nip93_to_nip68(private_key, RELAY_LIST, startindex))



import json
import os
from datetime import timedelta
from pathlib import Path
from typing import List

import dotenv
from nostr_sdk import Filter, Client, Alphabet, EventId, Event, PublicKey, Tag, Keys, nip04_decrypt, nip44_decrypt,  Metadata, Options, \
    Nip19Event, SingleLetterTag, RelayLimits, SecretKey, Connection, ConnectionTarget, \
    EventBuilder, Kind, ClientBuilder, SendEventOutput, NostrSigner

from nostr_dvm.utils.definitions import EventDefinitions, relay_timeout


async def get_event_by_id(event_id_str: str, client: Client, config=None) -> Event | None:
    split = event_id_str.split(":")
    if len(split) == 3:
        pk = PublicKey.parse(split[1])
        id_filter = Filter().author(pk).custom_tag(SingleLetterTag.lowercase(Alphabet.D), [split[2]])
        events = await client.fetch_events([id_filter], relay_timeout)
    else:
        event_id = EventId.parse(event_id_str)

        id_filter = Filter().id(event_id).limit(1)

        events = await client.fetch_events([id_filter], relay_timeout)

    if len(events.to_vec()) > 0:
        return events.to_vec()[0]
    else:
        print("Event not found")
        return None


async def get_events_async(client, filter, timeout):
    events = await client.fetch_events([filter], timedelta(seconds=timeout))
    return events.to_vec()


async def get_events_by_ids(event_ids, client: Client, config=None) -> List | None:
    search_ids = []
    events = []
    for event_id in event_ids:
        split = event_id.split(":")
        if len(split) == 3:
            pk = PublicKey.parse(split[1])
            id_filter = Filter().author(pk).custom_tag(SingleLetterTag.lowercase(Alphabet.D), [split[2]])
            events = await client.fetch_events([id_filter], relay_timeout)
        else:

            if str(event_id).startswith("nevent"):
                event_id = Nip19Event.from_bech32(event_id).event_id()
            elif str(event_id).startswith("nostr:nevent"):
                event_id = Nip19Event.from_nostr_uri(event_id).event_id()
            else:
                event_id = EventId.parse(event_id)
            search_ids.append(event_id)

            id_filter = Filter().ids(search_ids)
            events = await client.fetch_events([id_filter], relay_timeout)

    if len(events.to_vec()) > 0:
        return events.to_vec()
    else:
        return None


async def get_events_by_id(event_ids: list, client: Client, config=None) -> list[Event] | None:
    id_filter = Filter().ids(event_ids)
    # events = asyncio.run(get_events_async(client, id_filter, config.RELAY_TIMEOUT))
    events = await client.fetch_events([id_filter], relay_timeout)
    if len(events.to_vec()) > 0:
        return events.to_vec()
    else:
        return None


async def get_referenced_event_by_id(event_id, client, dvm_config, kinds) -> Event | None:
    if kinds is None:
        kinds = []
    if str(event_id).startswith("nevent"):
        event_id = Nip19Event.from_bech32(event_id).event_id()
    elif str(event_id).startswith("nostr:nevent"):
        event_id = Nip19Event.from_nostr_uri(event_id).event_id()
    else:
        event_id = EventId.parse(event_id)

    if len(kinds) > 0:
        job_id_filter = Filter().kinds(kinds).event(event_id).limit(1)
    else:
        job_id_filter = Filter().event(event_id).limit(1)
    events = await client.fetch_events([job_id_filter], relay_timeout)

    if len(events.to_vec()) > 0:
        return events.to_vec()[0]
    else:
        return None


async def get_inbox_relays(event_to_send: Event, client: Client, dvm_config):
    ptags = []
    for tag in event_to_send.tags().to_vec():
        if tag.as_vec()[0] == 'p':
            ptag = PublicKey.parse(tag.as_vec()[1])
            ptags.append(ptag)

    filter = Filter().kinds([EventDefinitions.KIND_RELAY_ANNOUNCEMENT]).authors(ptags)
    events = await client.fetch_events([filter], relay_timeout)
    if len(events.to_vec()) == 0:
        return []
    else:
        nip65event = events.to_vec()[0]
        relays = []
        for tag in nip65event.tags().to_vec():
            if ((tag.as_vec()[0] == 'r' and len(tag.as_vec()) == 2)
                    or ((tag.as_vec()[0] == 'r' and len(tag.as_vec()) == 3) and tag.as_vec()[2] == "read")):
                rtag = tag.as_vec()[1]
                if rtag.rstrip("/") not in dvm_config.AVOID_OUTBOX_RELAY_LIST:
                    if rtag.startswith("ws") and " " not in rtag:
                        relays.append(rtag)
        return relays


async def get_dm_relays(event_to_send: Event, client: Client, dvm_config):
    ptags = []
    for tag in event_to_send.tags().to_vec():
        if tag.as_vec()[0] == 'p':
            ptag = PublicKey.parse(tag.as_vec()[1])
            ptags.append(ptag)

    filter = Filter().kinds([Kind(10050)]).authors(ptags)
    events = await client.fetch_events([filter], relay_timeout)
    if len(events.to_vec()) == 0:
        return []
    else:
        nip65event = events.to_vec()[0]
        relays = []
        for tag in nip65event.tags().to_vec():
            if ((tag.as_vec()[0] == 'r' and len(tag.as_vec()) == 2)
                    or ((tag.as_vec()[0] == 'r' and len(tag.as_vec()) == 3) and tag.as_vec()[2] == "read")):
                rtag = tag.as_vec()[1]
                if rtag.rstrip("/") not in dvm_config.AVOID_OUTBOX_RELAY_LIST:
                    if rtag.startswith("ws") and " " not in rtag:
                        relays.append(rtag)
        return relays


async def get_main_relays(event_to_send: Event, client: Client, dvm_config):
    ptags = []
    for tag in event_to_send.tags().to_vec():
        if tag.as_vec()[0] == 'p':
            ptag = PublicKey.parse(tag.as_vec()[1])
            ptags.append(ptag)

    if len(await client.relays()) == 0:
        for relay in dvm_config.RELAY_LIST:
            await client.add_relay(relay)

    await client.connect()
    filter = Filter().kinds([EventDefinitions.KIND_FOLLOW_LIST]).authors(ptags)
    events = await client.fetch_events([filter], relay_timeout)
    if len(events.to_vec()) == 0:
        return []
    else:
        followlist = events.to_vec()[0]
        try:
            content = json.loads(followlist.content())
            relays = []
            for relay in content:
                if relay.rstrip("/") not in dvm_config.AVOID_OUTBOX_RELAY_LIST:
                    relays.append(relay)
            return relays
        except:
            return []


async def send_event_outbox(event: Event, client, dvm_config) -> SendEventOutput | None:
    # 1. OK, Let's overcomplicate things.
    # 2. If our event has a relays tag, we just send the event to these relay in the classical way.
    relays = []
    for tag in event.tags().to_vec():
        if tag.as_vec()[0] == 'relays':
            for index, param in enumerate(tag.as_vec()):
                if index != 0:
                    if tag.as_vec()[index].rstrip("/") not in dvm_config.AVOID_OUTBOX_RELAY_LIST:
                        try:
                            relays.append(tag.as_vec()[index])
                        except:
                            print("[" + dvm_config.NIP89.NAME + "] " + tag.as_vec()[
                                index] + " couldn't be added to outbox relays")
            break

    # 3. If we couldn't find relays, we look in the receivers inbox
    if len(relays) == 0:
        relays = await get_inbox_relays(event, client, dvm_config)

    # 4. If we don't find inbox relays (e.g. because the user didn't announce them, we just send to our default relays
    if len(relays) == 0:
        print("[" + dvm_config.NIP89.NAME + "] No Inbox found, replying to generic relays")
        relays = await get_main_relays(event, client, dvm_config)


    if len(relays) == 0:
        return
        # eventid = await send_event(event, client, dvm_config)
        # return eventid

    # 5. Otherwise, we create a new Outbox client with the inbox relays and send the event there
    relaylimits = RelayLimits.disable()
    connection = Connection().embedded_tor().target(ConnectionTarget.ONION)    
    
    # connection = Connection().addr("127.0.0.1:9050").target(ConnectionTarget.ONION)
    opts = Options().relay_limits(relaylimits).connection(connection)
    sk = SecretKey.parse(dvm_config.PRIVATE_KEY)
    keys = Keys.parse(sk.to_hex())
    outboxclient = ClientBuilder().signer(NostrSigner.keys(keys)).opts(opts).build()
    #print("[" + dvm_config.NIP89.NAME + "] Receiver Inbox relays: " + str(relays))

    for relay in relays[:5]:
        try:
            await outboxclient.add_relay(relay)
        except:
            print("[" + dvm_config.NIP89.NAME + "] " + relay + " couldn't be added to outbox relays")
    #
    await outboxclient.connect()
    try:
        #print("Connected, sending event")
        event_response = await outboxclient.send_event(event)

    except Exception as e:
        event_response = None
        print(e)

    # 5. Fallback, if we couldn't send the event to any relay, we try to send to generic relays instead.
    if event_response is None:
        relays = await get_main_relays(event, client, dvm_config)
        if len(relays) == 0:
            return None
        for relay in relays:
            await outboxclient.add_relay(relay)
        try:
            await outboxclient.connect()
            event_response = await outboxclient.send_event(event)
        except Exception as e:
            # Love yourself then.
            event_response = None
            print(e)


    await outboxclient.shutdown()

    return event_response


async def send_event(event: Event, client: Client, dvm_config, broadcast=False):
    try:
        relays = []
        for tag in event.tags().to_vec():
            if tag.as_vec()[0] == 'relays':
                for index, param in enumerate(tag.as_vec()):
                    if index != 0:
                        if tag.as_vec()[index].rstrip("/") not in dvm_config.AVOID_OUTBOX_RELAY_LIST:
                            try:
                                relays.append(tag.as_vec()[index])
                            except:
                                print("[" + dvm_config.NIP89.NAME + "] " + tag.as_vec()[
                                    index] + " couldn't be added to outbox relays")
                break

        relay_list = dvm_config.RELAY_LIST
        if broadcast:
            relay_list = dvm_config.ANNOUNCE_RELAY_LIST

        if len(relays) == 0:
            relays = relay_list

        for relay in relays:
            if relay not in dvm_config.RELAY_LIST:
                await client.add_relay(relay)

        await client.connect()

        try:
            response_status = await client.send_event(event)
        except Exception as e:
            print(e)
            response_status = None

        for relay in relays:
            if relay not in dvm_config.RELAY_LIST:
                await client.force_remove_relay(relay)
        return response_status
    except Exception as e:
        print(e)


def print_send_result(response_status):
    print("Success: " + str(response_status.success) + " Failed: " + str(response_status.failed) + " EventID: "
          + response_status.id.to_hex() + " / " + response_status.id.to_bech32())


def check_and_decrypt_tags(event, dvm_config):
    is_encrypted = False
    use_legacy_encryption = False

    try:
        p = ""
        for tag in event.tags().to_vec():
            if tag.as_vec()[0] == 'encrypted':
                is_encrypted = True
            elif tag.as_vec()[0] == 'p':
                p = tag.as_vec()[1]

        if is_encrypted:
            if p != dvm_config.PUBLIC_KEY:
                print("[" + dvm_config.NIP89.NAME + "] Task encrypted and not addressed to this DVM, "
                                                    "skipping..")
                return None, False

            elif p == dvm_config.PUBLIC_KEY:
                try:
                    tags_str = nip04_decrypt(Keys.parse(dvm_config.PRIVATE_KEY).secret_key(),
                                             event.author(), event.content())
                except:
                    try:
                        tags_str = nip44_decrypt(Keys.parse(dvm_config.PRIVATE_KEY).secret_key(),
                                                 event.author(), event.content())
                    except:
                        print("Wrong Nip44 Format")
                        return None, False
                    use_legacy_encryption = True

                params = json.loads(tags_str)
                params.append(Tag.parse(["p", p]).as_vec())
                params.append(Tag.parse(["encrypted"]).as_vec())
                event_as_json = json.loads(event.as_json())
                event_as_json['tags'] = params
                event_as_json['content'] = ""
                event = Event.from_json(json.dumps(event_as_json))
    except Exception as e:
        print(e)

    return event, use_legacy_encryption


def check_and_decrypt_own_tags(event, dvm_config):
    try:
        is_encrypted = False
        p = ""
        for tag in event.tags().to_vec():
            if tag.as_vec()[0] == 'encrypted':
                is_encrypted = True
            elif tag.as_vec()[0] == 'p':
                p = tag.as_vec()[1]

        if is_encrypted:
            if dvm_config.PUBLIC_KEY != event.author().to_hex():
                print("[" + dvm_config.NIP89.NAME + "] Task encrypted and not addressed to this DVM, "
                                                    "skipping..")
                return None

            elif event.author().to_hex() == dvm_config.PUBLIC_KEY:
                try:
                    tags_str = nip44_decrypt(Keys.parse(dvm_config.PRIVATE_KEY).secret_key(),
                                             PublicKey.parse(p), event.content())
                except:
                    tags_str = nip04_decrypt(Keys.parse(dvm_config.PRIVATE_KEY).secret_key(),
                                             PublicKey.parse(p), event.content())
                params = json.loads(tags_str)
                params.append(Tag.parse(["p", p]).as_vec())
                params.append(Tag.parse(["encrypted"]).as_vec())
                event_as_json = json.loads(event.as_json())
                event_as_json['tags'] = params
                event_as_json['content'] = ""
                event = Event.from_json(json.dumps(event_as_json))
    except Exception as e:
        print(e)

    return event


async def update_profile_lnaddress(private_key, dvm_config, lud16="", ):
    keys = Keys.parse(private_key)
    client = Client(NostrSigner.keys(keys))
    for relay in dvm_config.RELAY_LIST:
        await client.add_relay(relay)
    await client.connect()

    metadata = Metadata() \
        .set_lud16(lud16) \
        .set_nip05(lud16)

    await client.set_metadata(metadata)


async def update_profile(dvm_config, client, lud16="", broadcast=True):
    keys = Keys.parse(dvm_config.PRIVATE_KEY)
    try:
        nip89content = json.loads(dvm_config.NIP89.CONTENT)
        name = nip89content.get("name")
        about = nip89content.get("about")
        image = nip89content.get("picture")

        # Set metadata
        metadata = Metadata() \
            .set_name(name) \
            .set_display_name(name) \
            .set_about(about) \
            .set_picture(image) \
            .set_lud16(lud16) \
            .set_nip05(lud16)
        # .set_banner("https://example.com/banner.png") \


    except:
        metadata = Metadata() \
            .set_lud16(lud16) \
            .set_nip05(lud16)

    print("[" + dvm_config.NIP89.NAME + "] Setting profile metadata for " + keys.public_key().to_bech32() + "...")
    print(metadata.as_json())
    if broadcast:
       for relay in dvm_config.ANNOUNCE_RELAY_LIST:
           await client.add_relay(relay)
       await client.connect()

    return await client.set_metadata(metadata)


async def send_nip04_dm(client: Client, msg, receiver: PublicKey, dvm_config):
    keys = Keys.parse(dvm_config.PRIVATE_KEY)
    signer = NostrSigner.keys(keys)
    content = await signer.nip04_encrypt(receiver, msg)
    ptag = Tag.parse(["p", receiver.to_hex()])
    event = EventBuilder(Kind(4), content).tags([ptag]).sign_with_keys(Keys.parse(dvm_config.PRIVATE_KEY))
    await client.send_event(event)

    # relays = await get_dm_relays(event, client, dvm_config)
    #
    # outboxclient = Client(signer)
    # print("[" + dvm_config.NIP89.NAME + "] Receiver Inbox relays: " + str(relays))
    #
    # for relay in relays[:5]:
    #     try:
    #         await outboxclient.add_relay(relay)
    #     except:
    #         print("[" + dvm_config.NIP89.NAME + "] " + relay + " couldn't be added to outbox relays")
    # #
    # await outboxclient.connect()
    # try:
    #     print("Connected, sending event")
    #     event_id = await outboxclient.send_event(event)
    #     print(event_id.output)
    # except Exception as e:
    #     print(e)


def check_and_set_private_key(identifier):
    if not os.getenv("DVM_PRIVATE_KEY_" + identifier.upper()):
        pk = Keys.generate().secret_key().to_hex()
        add_pk_to_env_file("DVM_PRIVATE_KEY_" + identifier.upper(), pk)
        return pk
    else:
        return os.getenv("DVM_PRIVATE_KEY_" + identifier.upper())


def add_pk_to_env_file(dtag, oskey):
    env_path = Path('.env')
    if env_path.is_file():
        print(f'loading environment from {env_path.resolve()}')
        dotenv.load_dotenv(env_path, verbose=True, override=True)
        dotenv.set_key(env_path, dtag, oskey)



# Manual List of known relays we can not write to. It's optional, but this way we avoid
# even trying to send to them, avoiding potential errors or delays on the way.


AVOID_OUTBOX_RELAY_LIST = ["wss://nos.lol", "wss://nostr.fmt.wiz.biz", "wss://nostrelay.yeghro.site", "wss://nostr.wine",
                           "wss://filter.nostr.wine", "wss://relay.lightwork.space", "wss://onchain.pub",
                           "wss://nostr21.com", "wss://nostr.bitcoiner.social", "wss://nostr.orangepill.dev",
                           "wss://brb.io", "wss://relay.nostr.ch", "wss://nostr.rock", "wss://nostr.sandwich.farm",
                           "wss://nostr.onsats.org", "wss://nostr-pub.semisol.dev", "wss://no.str.cr",
                           "wss://nostr.zebedee.cloud", "wss://blg.nostr.sx", "wss://relay.nostr.ai",
                           "wss://nostr.600.wtf", "ws://umbrel.local:4848", "wss://192.168.1.52:5051",
                           "wss://nostrvista.aaroniumii.com", "wss://https//nostr.einundzwanzig.space",
                           "wss://nostr.localhost.re", "wss://shopstr.store", "wss://th1.nostr.earnkrub.xyz",
                           "wss://relay.lnpay.me", "wss://relay.snort.social", "wss://relay.minds.com/nostr/v1/ws",
                           "ws://elitedesk:4848", "wss://wot.nostr.net", "wss://blg.nostr.sx",
                           "wss://nostr-pub.semisol.dev", "wss://mostr.mostr.pub", "wss://relay.mostr.pub",
                           "wss://minds.com", "wss://nostr.leximaster.com", "wss://th2.nostr.earnkrub.xyz",
                           "wss://relay.zerosatoshi.xyz", "wss://bouncer.minibolt.info",
                           "wss://yabu.me", "wss://relay.yozora.world", "wss://filter.nostr.wine/?global=all",
                           "wss://eden.nostr.land", "wss://relay.otherstuff.fyi",
                           "wss://relay.orangepill.ovh", "wss://nostr.jcloud.es", "wss://af.purplerelay.com",
                           "wss://za.purplerelay.com", "ws://192.168.18.7:7777", "wss://nostr.fediverse.jp",
                           "wss://relay.nostrich.land", "wss://relay.nostrplebs.com", "wss://relay.nostrich.land",
                           "ws://elitedesk.local:4848", "wss://nrelay-jp.c-stellar.net", "wss://blastr.f7z.xyz",
                           "wss://rss.nos.social", "wss://atlas.nostr.land", "wss://puravida.nostr.land",
                           "wss://nostr.inosta.cc", "wss://relay-jp.nostr.wirednet.jp",
                           "wss://relay.orangepill.dev", "wss://no.str.cr", "wss://nostr.milou.lol",
                           "wss://relay.nostr.com.au", "wss://nostrelites.com", "wss://sfr0.nostr1.com",
                           "wss://puravida.nostr.land", "wss://atlas.nostr.land", "wss://nostr-pub.wellorder.net",
                           "wss://relay.current.fyi", "wss://nfrelay.app",
                           "wss://nostr.thesamecat.io", "wss://nostr.plebchain.org", "wss://relay.noswhere.com",
                           "wss://nostr.uselessshit.co", "wss://tictac.nostr1.com", "wss://abcdefg20240104205400.xyz/v1",
                           "wss://bitcoiner.social", "wss://relay.stoner.com", "wss://nostr.l00p.org",
                           "wss://relay.nostr.ro", "wss://nostr.kollider.xyz",
                           "wss://relay.valera.co", "wss://relay.austrich.net", "wss://relay.nostrich.de",
                           "wss://nostr.azte.co", "wss://nostr-relay.schnitzel.world",
                           "wss://relay.nostriches.org", "wss://happytavern.co", "wss://onlynotes.lol",
                           "wss://offchain.pub", "wss://purplepag.es", "wss://relay.plebstr.com",
                           "wss://poster.place/relay", "wss://relayable.org", "wss://bbb.santos.lol",
                           "wss://relay.bitheaven.social", "wss://theforest.nostr1.com", "wss://at.nostrworks.com",
                           "wss://relay.nostrati.com", "wss://purplerelay.com", "wss://hist.nostr.land",
                           "wss://creatr.nostr.wine", "ws://localhost:4869", "wss://pleb.cloud",
                           "wss://pyramid.fiatjaf.com", "wss://relay.nos.social", "wss://nostr.thank.eu",
                           "wss://inbox.nostr.wine", "wss://relay.pleb.to", "wss://welcome.nostr.wine",
                           "wss://relay.nostrview.com", "wss://nostr.land", "wss://eu.purplerelay.com",
                           "wss://xmr.usenostr.org", "wss://nostr-relay.app", "ws://umbrel:4848", "wss://umbrel:4848",
                           "wss://fiatjaf.com", "wss://nostr-relay.wlvs.space", "wss://relayer.fiatjaf.com",
                           "wss://nostr.yuv.al", "wss://relay.nostr.band", "wss://nostr.massmux.com",
                           "wss://nostr-01.bolt.observer", "wss://nostr1.tunnelsats.com", "wss://relay.nostr.ch",
                           "wss://relay.nostr.io", "wss://nostr.thank.eu", "wss://nostr.bitcoinplebs.de",
                           "wss://adult.18plus.social", "wss://bostr.online", "wss://relay.current.fyi",
                           "wss://nosdrive.app/relay", "wss://studio314.nostr1.com", "wss://relay.nostrbr.online",
                           "wss://relay.0xchat.com", "wss://lightning.benchodroff.com/nostrclient/api/v1/relay",
                           "wss://relay.nostreggs.io", "wss://relay.blackbyte.nl", "ws://localhost:8080",
                           "wss://127.0.0.1:4869", "wss://sendit.nosflare.io",
                           "wss://astral.ninja", "wss://nostr.libertasprimordium.com", "wss://relay.shitforce.one",
                           "wss://nostr.cro.social", "wss://datagrave.wild-vibes.ts.net/nostr", "wss://nostr01.sharkshake.net",
                           "wss://relay.nostreggs.io", "wss://nostr.rocks", "wss://groups.0xchat.com",
                           "wss://bostr.lecturify.net", "wss://dave.st.germa.in/nostr",
                           "wss://nostr.social", "wss://i.nostr.build", "wss://invillage-outvillage.com",
                           "wss://teemie1-relay.duckdns.org", "wss://newperspectives.duckdns.org",
                           "wss://nostrs.build", "wss://relay.hllo.live", "wss://relay-pub.deschooling.us",
                           "wss://nostr.sandwich.farm", "wss://nostr.lol", "wss://nostr.developer.li",
                           "wss://paid.spore.ws", "ws://relay.damus.io", "wss://taipei.scarab", "wss://taipei.scarab.im",
                           "ws://ofotwjuiv7t6q4azt2fjx3qo7esglmxdeqmh2qvdsdnxw5eqgza24iyd.onion", "wss://r.kojira.io",
                           "wss://nostr-relay.h3z.jp", "wss://relay.yozora.world",
                           "wss://nostr.0xtr.dev", "wss://purplepeg.es", "wss://nostr.mutinywallet.com",
                           "wss://nostr.zebedee.cloud", "wss://relay.wikifreedia.xyz", "wss://relay.exit.pub",
                           "wss://njump.mecheck", "wss://relay.nostr.band", "wss://nostr.kollider.xyz",
                           "wss://fog.dedyn.io", "wss://relay.current.fyi",
                           "wss://momostr.pink", "wss://nostr.bitcoinlighthouse.de", "wss://140.f7z.io",
                           "wss://relay.nostrcheck.me", "wss://relay.mostr.pub", "wss://purplepag.es",
                           "wss://nostr.bitcoiner.social", "wss://relay.galtsgulch.cc",
                           "ws://oxtrdevav64z64yb7x6rjg4ntzqjhedm5b5zjqulugknhzr46ny2qbad.onion",
                           "wss://relay.f7z.io", "wss://nostr-relay.h3z.jp", "wss://nfrelay.app",
                           "wss://r.kojira.io", "wss://jp-relay-nostr.invr.chat",
                           "wss://nostream-production-5895.up.railway.app",
                           "ws://127.0.0.1:4869", "wss://sign.siamstr.com", "wss://relay.hash.stream",
                           "wss://pablof7z.nostr1.com", "wss://nostr.beckmeyer.us", "wss://pow.hzrd149.com",
                           "wss://relay.nostrss.re", "wss://relay.nostr.bg", "ws://bugman.mguy.net:4848",
                           "wss://nostr-2.zebedee.cloud", "wss://nostr.lorentz.is", "wss://zap.nostr1.com",
                           "wss://onchain.pub", "wss://relay.nostr.info", "wss://relay.chicagoplebs.com",
                           "wss://relay.current.fyi", "wss://relay.stemstr.app", "wss://nostr.zenon.info",
                           "ws://localhost:7777", "wss://nostr.fmt.wiz.biz", "wss://nostrich.friendship.tw",
                           "wss://public.relaying.io", "wss://relay.me3d.app", "wss://dreamofthe90s.nostr1.com",
                           "wss://lnbits.eldamar.icu/nostrrelay/relay", "wss://nostr.olwe.link",
                           "wss://nostr.cheeserobot.org", "wss://nostr-relay.nokotaro.com", "wss://premium.primal.net"





                           ]



import base64
import hashlib

from nostr_sdk import EventBuilder, Tag, Kind, Keys


def sha256sum(filename):
    with open(filename, 'rb', buffering=0) as f:
        return hashlib.file_digest(f, 'sha256').hexdigest()


async def generate_nip98_header(pkeys_hex, url="", kind="POST", filepath=""):
    """
    Generates a NIP-98 authentication header for a Nostr-based server. This header is
    intended to be used for authenticating requests by signing events with specific
    keys and including necessary metadata tags. The resulting header is encoded into
    a format suitable for HTTP authorization headers.

    :param pkeys_hex: The private keys in hexadecimal format, used for signing the
        event.
    :type pkeys_hex: str
    :param url: The URL that the NIP-98 header should apply to. Defaults to an empty
        string.
    :type url: str, optional
    :param kind: The HTTP method type for which the header is generated (e.g.,
        "POST"). Defaults to "POST".
    :type kind: str, optional
    :param filepath: The path to a file whose content will be hashed and included in
        the payload as a tag if the method is "POST". Defaults to an empty string.
    :type filepath: str, optional
    :return: The generated NIP-98 header in the format "Nostr <Base64 encoded event>".
    :rtype: str
    """
    keys = Keys.parse(pkeys_hex)
    utag = Tag.parse(["u", url])
    methodtag = Tag.parse(["method", kind])
    tags = [utag, methodtag]
    if kind == "POST":
        payloadtag = Tag.parse(["payload", sha256sum(filepath)])
        tags.append(payloadtag)
    eb = EventBuilder(Kind(27235), "").tags(tags)
    event  = eb.sign_with_keys(keys)

    encoded_nip98_event = base64.b64encode(event.as_json().encode('utf-8')).decode('utf-8')

    return "Nostr " + encoded_nip98_event



from nostr_sdk import Tag, Keys, EventBuilder, Kind

from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.nostr_utils import send_event
from nostr_dvm.utils.print_utils import bcolors


async def announce_dm_relays(dvm_config, client):
    tags = []

    for relay in dvm_config.RELAY_LIST:
        r_tag = Tag.parse(["r", relay])
        tags.append(r_tag)

    keys = Keys.parse(dvm_config.NIP89.PK)
    content = ""

    event = EventBuilder(Kind(10050), content).tags(tags).sign_with_keys(keys)
    response_status = await send_event(event, client=client, dvm_config=dvm_config, broadcast=True)
    if response_status is not None:
        print(
            bcolors.BLUE + "[" + dvm_config.NIP89.NAME + "] Announced DM relays for " + dvm_config.NIP89.NAME + ". Success: " + str(
                response_status.success) + " Failed: " + str(response_status.failed) + " EventID: "
            + response_status.id.to_hex() + " / " + response_status.id.to_bech32())
    else:
        print(
            bcolors.RED + "[" + dvm_config.NIP89.NAME + "] Could not announce DM relays for " + dvm_config.NIP89.NAME + bcolors.ENDC)


async def nip65_announce_relays(dvm_config, client):
    # todo we might want to call the dm relays seperately but for now we do it together with the inbox relays
    await announce_dm_relays(dvm_config, client)

    tags = []

    for relay in dvm_config.RELAY_LIST:
        r_tag = Tag.parse(["r", relay])
        tags.append(r_tag)

    keys = Keys.parse(dvm_config.NIP89.PK)
    content = ""

    event = EventBuilder(EventDefinitions.KIND_RELAY_ANNOUNCEMENT, content).tags(tags).sign_with_keys(keys)
    response_status = await send_event(event, client=client, dvm_config=dvm_config, broadcast=True)
    if response_status is not None:
        print(
            bcolors.BLUE + "[" + dvm_config.NIP89.NAME + "] Announced NIP 65 for " + dvm_config.NIP89.NAME + ". Success: " + str(
                response_status.success) + " Failed: " + str(response_status.failed) + " EventID: "
            + response_status.id.to_hex() + " / " + response_status.id.to_bech32())
    else:
        print(
            bcolors.RED + "[" + dvm_config.NIP89.NAME + "] Could not announce NIP 65 for " + dvm_config.NIP89.NAME + bcolors.ENDC)
